#!/usr/bin/env python
# coding: utf-8

"""
This shows how message filtering works.
"""

import time
import datetime
import os
import can
from html.parser import HTMLParser

t_last_amps=0.0

#isetlist=[20,21,22,23,24,25,26,27,28,29]
isetlist=[0.5,0.5,0.5,0.5,0.7,0.9,1.1,1.4,2.0,2.0,2.0,2.0,1.4,1.1,1.1,0.7,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5]
isetindex=0

vitmcount=0

g_ghi="NI"
g_ghy="NI"
g_gti="NI"
g_gty="NI"
g_next=0

# create a subclass and override the handler methods
class MyHTMLParser(HTMLParser):
    def handle_data(self, data):
        global g_ghi,g_ghy,g_gti,g_gty,g_next
        data = data.strip()
        #print ("Encountered some data :", len(data),data)

        if data=="Global Horiz. Today":
            #print("Waiting for GHI")
            g_next=1

        elif data=="Global Horiz. Yesterday":
            #print("Waiting for GHY")
            g_next=2

        elif data=="Global Tilt Today":
            #print("Waiting for GTI")
            g_next=3

        elif data=="Global Tilt Yesterday":
            #print("Waiting for GTY")
            g_next=4

        elif len(data):
            if g_next==1:
                g_ghi=data
            elif g_next==2:
                g_ghy=data
            elif g_next==3:
                g_gti=data
            elif g_next==4:
                g_gty=data
            g_next=0


g_lv=0.0
g_hv=0.0
g_temp=0.0 
g_lv = 0.0
g_wh = 0.0
g_ihv = 0.0 
g_vpv = 0.0 
g_tf = 0.0
g_tt = 0.0


class realstat:
    count=0
    first_timestamp=0.0
    last_timestamp=0.0
    mutex_timestamp=0.0
    last_value=0.0
    accumulator=0.0
    integrator=0.0
    min_value=0.0
    max_value=0.0
    last_value=0.0

    def __init__( self ):
        self.count=0

    def add_value( self, value , timestamp ):
        mutex_timestamp=timestamp
        if self.count==0:
            self.first_timestamp=timestamp
            self.min_value=value
            self.max_value=value
        else:
            if value<self.min_value:
                self.min_value=value
            if value>self.max_value:
                self.max_value=value

            if (timestamp-self.last_timestamp) < 2.0:
                self.integrator+=(timestamp-self.last_timestamp)*value

        self.last_value=value
        self.accumulator+=value
        self.count+=1
        self.last_timestamp=timestamp

    def reset(self):
        self.mutex_timestamp=0.0
        self.last_value=0.0
        self.count=0
        self.accumulator=0.0
        self.integrator=0.0
        self.first_timestamp=0.0
        self.last_timestamp=0.0

class stat_tracker():
    cell_voltage=[]
    cell_temperature=[]
    last_timestamp=0.0

    def __init__(self):
        for count in range(96):
            self.cell_voltage.append(  realstat()  )
        for count in range(9):
            self.cell_temperature.append(  realstat()  )
        self.wh_regen=realstat()
        self.wh_drive=realstat()
        self.wh_net=realstat()
        self.wh_ac=realstat()
        self.wh_hv=realstat()
        self.wh_lv=realstat()
        self.wh_pv=realstat()
        self.a_pv=realstat()
        self.inlet_temperature=realstat()
        self.outlet_temperature=realstat()
        self.pack_voltage=realstat()
        self.twelve_voltage=realstat()
        self.pv_voltage=realstat()
        self.fet_temperature=realstat()
        self.trans_temperature=realstat()
        self.wh_pv_nv=realstat()
        self.msfc=realstat()
        self.buck_current=realstat()
        self.buck_voltage=realstat()
        self.buck_temperature=realstat()
        self.pcdb_vhv=realstat()

    def reset_all(self):
        for cell in self.cell_voltage:
            cell.reset()
        for cell in self.cell_temperature:
            cell.reset()
        self.wh_regen.reset()
        self.wh_drive.reset()
        self.wh_net.reset()
        self.wh_ac.reset()
        self.wh_hv.reset()
        self.wh_lv.reset()
        self.wh_pv.reset()
        self.a_pv.reset()
        self.inlet_temperature.reset()
        self.outlet_temperature.reset()
        self.pack_voltage.reset()
        self.twelve_voltage.reset()
        self.pv_voltage.reset()
        self.fet_temperature.reset()
        self.trans_temperature.reset()
        self.wh_pv_nv.reset()
        self.msfc.reset()
        self.buck_current.reset()
        self.buck_voltage.reset()
        self.buck_temperature.reset()
        self.pcdb_vhv.reset()

    def min_cell(self):
        min_v=10.0
        for cell in self.cell_voltage:
            if cell.last_value<min_v:
                min_v=cell.last_value
        print ("min_cell_v ", min_v)
        return min_v

    def max_cell(self):
        max_v=-10.0
        for cell in self.cell_voltage:
            if cell.last_value>max_v:
                max_v=cell.last_value
        print ("max_cell_v ", max_v)
        return max_v

    def avg_cell(self):
        sum_v=0.0
        sum_n=0
        for cell in self.cell_voltage:
            sum_v += cell.last_value
            sum_n += 1
        return sum_v/sum_n

    def min_cell_temperature(self):
        min_v=200.0
        for cell in self.cell_temperature:
            if cell.last_value<min_v:
                min_v=cell.last_value
        return min_v

    def max_cell_temperature(self):
        max_v=-10.0
        for cell in self.cell_temperature:
            if cell.last_value>max_v:
                max_v=cell.last_value
        return max_v

    def avg_cell_temperature(self):
        sum_v=0.0
        sum_n=0
        for cell in self.cell_temperature:
            sum_v += cell.last_value
            sum_n += 1
        return sum_v/sum_n

    def sum_cell(self):
        sum_v=0.0
        for cell in self.cell_voltage:
            sum_v += cell.last_value
        return sum_v

    def print_cells(self):
        total=0.0
        print ("cells ",)
        for cell in self.cell_voltage:
            print( cell.last_value,)
            total+=cell.last_value
        print (" sum ", total)

g_stat_tracker = stat_tracker()

def rxcallback( arg ):
    global t_last_amps    
    global isetlist
    global isetindex
    global vitmcount
    global g_lv
    global g_hv
    global g_temp
    global g_wh
    global g_ihv
    global g_tf
    global g_tt
    global g_vpv
    global g_stat_tracker

    pid= arg.arbitration_id
    data = arg.data
    g_stat_tracker.last_timestamp=arg.timestamp
    print_raw=0
    if print_raw:
        print ("{:.6f}".format(arg.timestamp) ,)
        print ("{:03x}".format(pid) , )
        print (" ".join('{:02x}'.format((c)) for c in arg.data ), )
   
    if pid==0x30A:  # charger AC power
      #  print "30A ", " ".join('{:02x}'.format((c)) for c in arg.data ), 

        acc= ((data[0]<<4) + (data[1]>>4))/5.0
        acv= (((data[1]<<4) + (data[2]>>4))&0xff)*2.0

        g_stat_tracker.wh_ac.add_value( acc*acv , arg.timestamp )
        if print_raw:
            print (acc, acv)
   #     for c in arg.data:
  #         print c
    elif pid==0x210:  # old battery readings guess
        hvv= (((data[0]<<4) + (data[1]>>1))&0xfff)/8.0
        hvc1= (((data[1]<<8) + (data[2]>>0))&0x3ff)/10.0
        hvc2= (((data[2]<<7) + (data[4]>>1))&0xfff)/20.0
        if print_raw:
            print (hvv, hvc1, hvc2 )

    elif pid==0x264:  # minutes since last full charge 
        mins= (((data[0]<<8) + (data[1]>>0))&0xfff)/2.0
        g_stat_tracker.msfc.add_value( mins , arg.timestamp )
        if print_raw:
            print( mins)

    elif pid==0x216:  # my battery readings 
        hvv= (((data[1]<<8) + (data[2]>>0))&0xfff)/8.0
        g_hv=hvv
        hvch= (((data[3]<<8) + (data[4]>>0))&0xffff)*1.0
        hvcl= (((data[5]<<8) + (data[6]>>0))&0xffff)*1.0
        if hvch >= 32768 :
            hvch-=65536
        if hvcl >= 32768 : # quick hack to process 2's complement 
            hvcl-=65536
        hvch = hvch/50.0
        hvcl = hvcl/50.0
        if abs(hvcl) > 15.0 :
            bpow= hvv*hvch
        else:
            bpow= hvv*hvcl
        if bpow>0.01:
            g_stat_tracker.wh_regen.add_value( bpow , arg.timestamp )
            g_stat_tracker.wh_drive.add_value( 0.0 , arg.timestamp )
        else:
            g_stat_tracker.wh_regen.add_value( 0.0 , arg.timestamp )
            g_stat_tracker.wh_drive.add_value( -bpow , arg.timestamp )
        g_stat_tracker.wh_net.add_value( bpow , arg.timestamp )
        g_stat_tracker.pack_voltage.add_value( hvv , arg.timestamp )

        if print_raw:
            print( hvv, hvch, hvcl, bpow)
        vitmcount = vitmcount+1
        if vitmcount>=20:
        #    print( "Battery Power ", "{:6.1f}".format(bpow))
            vitmcount=0

    elif pid==0x460:  # coolant temp
        tin = (((data[0]<<8) + (data[1]>>0))&0x7ff)/8.0 -40.0 # assuming heater on the inlet
        tout= (((data[2]<<8) + (data[3]>>0))&0x7ff)/8.0 -40.0
        g_stat_tracker.inlet_temperature.add_value(tin,arg.timestamp)
        g_stat_tracker.outlet_temperature.add_value(tout,arg.timestamp)
        if print_raw:
            print( tin, tout)


    elif pid==0x1d6:  # apu status, other bus
        vo = (((data[2]))&0xff)/12.7
        t1 = (((data[3]))&0xff)/1.0 -40
        t2 = (((data[4]))&0xff)/1.0 -40
        io = (((data[5]))&0xff)/1.0

        po = vo*io
        if print_raw:
            print( vo, t1, t2, io, po)

    elif pid==0x302:  # module temp
        t1 = (((data[1]))&0xff)/2.0 -40
        t2 = (((data[2]))&0xff)/2.0 -40
        t3 = (((data[3]))&0xff)/2.0 -40
        t4 = (((data[4]))&0xff)/2.0 -40
        t5 = (((data[5]))&0xff)/2.0 -40
        t6 = (((data[6]))&0xff)/2.0 -40
        if data[7]==0x30:
          if print_raw:
            print( t1, t2, t3)
          g_stat_tracker.cell_temperature[ 6 ].add_value(t1,arg.timestamp)
          g_stat_tracker.cell_temperature[ 7 ].add_value(t2,arg.timestamp)
          g_stat_tracker.cell_temperature[ 8 ].add_value(t3,arg.timestamp)
        else:
          if print_raw:
            print( t1, t2, t3, t4, t5, t6 )
          g_stat_tracker.cell_temperature[ 0 ].add_value(t1,arg.timestamp)
          g_stat_tracker.cell_temperature[ 1 ].add_value(t2,arg.timestamp)
          g_stat_tracker.cell_temperature[ 2 ].add_value(t3,arg.timestamp)
          g_stat_tracker.cell_temperature[ 3 ].add_value(t4,arg.timestamp)
          g_stat_tracker.cell_temperature[ 4 ].add_value(t5,arg.timestamp)
          g_stat_tracker.cell_temperature[ 5 ].add_value(t6,arg.timestamp)
          g_temp = (t1+t2+t3+t4+t5+t6) / 6.0 

    elif pid>=0x200  and pid<=0x206 :  # cell voltage
        cell1 = (((data[0]<<7) + (data[1]>>1))&0xfff)/800.0
        cell2 = (((data[2]<<7) + (data[3]>>1))&0xfff)/800.0
        cell3 = (((data[4]<<4) + (data[5]>>4))&0xfff)/800.0
        pkt = (((data[5]<<3) + (data[6]>>5))&0x7f)
        cell_num=int(((pid-0x200)/2)*24 + pkt*3 )
        if print_raw:
            print ("cell " +  str(cell_num))
        g_stat_tracker.cell_voltage[cell_num + 0 ].add_value(cell1,arg.timestamp)
        g_stat_tracker.cell_voltage[cell_num + 1 ].add_value(cell2,arg.timestamp)
        g_stat_tracker.cell_voltage[cell_num + 2 ].add_value(cell3,arg.timestamp)
        if print_raw:
            print (cell1, cell2, cell3, pkt)

    elif pid==0x212:   # charger DC power 
       # print "212 ", " ".join('{:02x}'.format((c)) for c in arg.data ), 
        dcbin=data[0]*256 + data[1]
        hvc= ((data[0]<<5) + (data[1]>>3))/20.0
        hvv= (((data[1]<<7) + (data[2]>>1))&0x3ff)/2.0
        lvc= (((data[2]<<7) + (data[3]>>1))&0xff)/5.0
        lvv= (((data[3]<<7) + (data[4]>>1))&0xff)/10.0
        g_lv = lvv
        g_stat_tracker.wh_hv.add_value( hvc*hvv , arg.timestamp )
        g_stat_tracker.wh_lv.add_value( lvc*lvv , arg.timestamp )
        g_stat_tracker.twelve_voltage.add_value( lvv , arg.timestamp )
        if print_raw:
            print( hvc, hvv, lvc, lvv)
    
    elif pid==0x350:   # PV DC power 
       # print "212 ", " ".join('{:02x}'.format((c)) for c in arg.data ), 
        #dcbin=data[0]*256 + data[1]
        wh=0.0 #FIXME
        vhvi= ((data[0]<<8) + (data[1]>>0))
        if vhvi>= 32768 : # quick hack to process 2's complement 
            vhvi-=65536
        vhv=vhvi/1000.0*20.0
        ihvi= ((data[2]<<8) + (data[3]>>0))
        if ihvi>= 32768 : # quick hack to process 2's complement 
            ihvi-=65536
        ihv=ihvi/1000.0
        vpv= ((data[6]<<8) + (data[7]>>0))/1000.0

        tf=data[4]*1.0
        tt=data[5]*1.0
        g_stat_tracker.pcdb_vhv.add_value( vhv  , arg.timestamp )
        g_stat_tracker.wh_pv_nv.add_value( wh  , arg.timestamp )
        g_stat_tracker.wh_pv.add_value(ihv* vhv , arg.timestamp )
        g_stat_tracker.a_pv.add_value(ihv , arg.timestamp )
        g_stat_tracker.fet_temperature.add_value(tf , arg.timestamp )
        g_stat_tracker.trans_temperature.add_value(tt , arg.timestamp )
        g_stat_tracker.pv_voltage.add_value( vpv , arg.timestamp )
        # using VITM voltage is ok, since converter will not run without receiving VITM voltage itself

        g_wh = wh
        g_ihv = ihv 
        g_vpv = vpv 
        g_tf = tf 
        g_tt = tt
        if print_raw:
            print( wh, ihv, tf, tt, vpv)

    elif pid==0x352:   # PV DC power 
       # print "212 ", " ".join('{:02x}'.format((c)) for c in arg.data ), 
        #dcbin=data[0]*256 + data[1]
        wh=0.0 #FIXME
        ibucki= ((data[2]<<8) + (data[3]>>0))
        if ibucki>= 32768 : # quick hack to process 2's complement 
            ibucki-=65536
        ibuck=ibucki/1000.0
        ibuckv= ((data[0]<<8) + (data[1]>>0))
        if ibuckv>= 32768 : # quick hack to process 2's complement 
            ibuckv-=65536
        vbuck=ibuckv/1000.0
        vmppt= ((data[6]<<8) + (data[7]>>0))/1000.0

        tb=data[4]*1.0
        g_stat_tracker.buck_temperature.add_value(tb , arg.timestamp )
        g_stat_tracker.buck_voltage.add_value( vbuck , arg.timestamp )
        g_stat_tracker.buck_current.add_value( ibuck , arg.timestamp )
        # using VITM voltage is ok, since converter will not run without receiving VITM voltage itself


    elif pid==0x304:   # charger output setting 
        #print "304 ", " ".join('{:02x}'.format((c)) for c in arg.data ), 
        #print t_last_amps-arg.timestamp, 
        t_last_amps=arg.timestamp

        cc= ((data[1]) )/20.0
        cv= (((data[2]<<8) + (data[3]>>0))&0x3ff)/2.0
        ilim=50  #15 too low 

        if isetindex>=len(isetlist):
            isetindex=0
        ilim=isetlist[isetindex]
        isetindex+=1

#        if isetindex<10:
#            ilim=isetindex
#        else:
#            ilim=20-isetindex


        if data[1]>ilim:
            cs=ilim
        else:
            cs=data[1]

        cs=int( data[1]*ilim) 
        if cs>250:
            cs=250
        #cs=data[1]/1 
        if print_raw:
            print( cc, cv, "old", data[1], "new",cs)
        msg = can.Message(arbitration_id=0x304,
                      data=[data[0],cs,data[2],data[3] ],  # half current
                      extended_id=False) 
        if 0:
          try:
           bus.send(msg)
          #  print("Message sent on {}".format(bus.channel_info))
          except can.CanError:
           print("Message NOT sent")

    else:   # undefined PIDs
        if print_raw:
            print()



if __name__ == '__main__':
    bus = can.interface.Bus(bustype='socketcan',
                            channel='can0',
                            receive_own_messages=False)

    can_filters = [{"can_id": 1, "can_mask": 0xf, "extended": True}]
   # bus.set_filters(can_filters)
  #  notifier = can.Notifier(bus, [rxcallback])
    #notifier = can.Notifier(bus, [can.Printer()])
  #  bus.send(can.Message(arbitration_id=1, extended_id=True))
  #  bus.send(can.Message(arbitration_id=2, extended_id=True))
  #  bus.send(can.Message(arbitration_id=1, extended_id=False))
    rxcallback.t_last_amps = 0

    mylistener = can.Listener()
    mylistener.on_message_received = rxcallback
    mylistener.bus=bus
    notifier = can.Notifier(bus, [mylistener])
    upload=0
    log=0
    check_snl=0
    d_wh_drive=0.0
    d_wh_regen=0.0
    d_wh_net=0.0
    d_wh_ac=0.0
    d_wh_hv=0.0
    d_wh_lv=0.0
    d_wh_pv=0.0
    d_s_pv=0.0
    d_s_drive=0.0
    d_s_ac=0.0
    d_s_wake=0.0
    d_m_dist=0.0
    d_v_twelve=0.0
    d_v_stack=0.0
    d_v_sum=0.0
    d_v_min=0.0
    d_v_max=0.0
    d_t_ci=88.8
    d_t_co=88.8
    d_t_avg=88.8
    d_t_min=88.8
    d_t_max=88.8
    reset_daily=0
    reset_stats=0
    string_daily=0
    lastlooptimedate=datetime.datetime.now()
    while (1):
        tsession= g_stat_tracker.wh_net.last_timestamp-g_stat_tracker.wh_net.first_timestamp


        if (time.time()-g_stat_tracker.wh_net.last_timestamp)>10.0:
            if g_stat_tracker.wh_net.last_timestamp !=0.0:
               print("end of session, duration ", tsession)
               # string_daily=1 post at end of session
               d_wh_ac+=g_stat_tracker.wh_ac.integrator/3600.0
               d_wh_hv+=g_stat_tracker.wh_hv.integrator/3600.0
               d_wh_lv+=g_stat_tracker.wh_lv.integrator/3600.0
               d_wh_pv+=g_stat_tracker.wh_pv.integrator/3600.0
               d_t_ci=g_stat_tracker.inlet_temperature.last_value
               d_t_co=g_stat_tracker.outlet_temperature.last_value
               d_v_twelve=g_stat_tracker.twelve_voltage.last_value
               d_v_stack=g_stat_tracker.pack_voltage.last_value
               d_v_sum=g_stat_tracker.sum_cell()
               d_v_min=g_stat_tracker.min_cell()
               d_v_max=g_stat_tracker.max_cell()
               d_t_avg=g_stat_tracker.avg_cell_temperature()
               d_t_min=g_stat_tracker.min_cell_temperature()
               d_t_max=g_stat_tracker.max_cell_temperature()
               # session classifier
               if g_stat_tracker.wh_ac.integrator > 100.0:
                   d_s_ac+=tsession
               elif g_stat_tracker.wh_pv.integrator > 100.0:  # ok with not pv in motion for now
                   d_s_pv+=tsession
               elif g_stat_tracker.wh_drive.integrator > 1000.0: # 200W * 5s = 1000J
                   d_s_drive+=tsession
                   d_wh_drive+=g_stat_tracker.wh_drive.integrator/3600.0
                   d_wh_regen+=g_stat_tracker.wh_regen.integrator/3600.0
                   d_wh_net+=g_stat_tracker.wh_net.integrator/3600.0
               else:
                   d_s_wake+=tsession
                   d_wh_drive+=g_stat_tracker.wh_drive.integrator/3600.0
                   d_wh_regen+=g_stat_tracker.wh_regen.integrator/3600.0
                   d_wh_net+=g_stat_tracker.wh_net.integrator/3600.0

               log=1
               reset_stats=1
        #logstr += (",%6.1f,%5.3f,%4.1f,%4.1f,%4.1f" % ( g_wh, g_ihv, g_tf, g_tt, g_vpv) )
        #print (logstr)
        #g_stat_tracker.print_cells()
        g_lv=0.0
        g_hv=0.0
        g_temp=0.0


        thislooptimedate=datetime.datetime.now()
       # if thislooptimedate.hour != lastlooptimedate.hour:
        if thislooptimedate.minute != lastlooptimedate.minute:
            lmin=thislooptimedate.minute
            lhour=thislooptimedate.hour
            log=1
            #print ("Mark! ", lhour, lmin)
            if (lhour>5) & (lhour<22):
                if lmin==2:
                    check_snl=1
                if lmin==3:
                    check_snl=2

            if lmin==3:
                if lhour==21:
                    string_daily=1
                    upload=1
                    reset_daily=1


        if check_snl==1:
            check_snl=0
            os.system("curl -k --max-time 45 --max-filesize 1000000 -A changethis@realsolarcars.com https://photovoltaics.sandia.gov/weather/Weather.htm --output /dev/shm/Weather.htm & ")

        if check_snl==2:
            check_snl=0
            try:
                # instantiate the parser and fed it some HTML
                parser = MyHTMLParser()
                with open("/dev/shm/Weather.htm") as f:
                  for line in f:
                    #print line,
                    #parser.feed(line.encode('utf8', 'replace') )
                    #parser.feed( filter(onlyascii, line) )
                    parser.feed( line )
            except:
                print("SNL Parse Exception")
                g_ghi="PE" 
                g_ghy="PE"
                g_gti="PE" 
                g_gty="PE" 

            csvstr="%s,%s,%s,%s" % ( g_ghi, g_ghy, g_gti, g_gty )
            print("SNL: ",csvstr)

        if log:
            log=0
          #  logstr =  datetime.datetime.now().strftime("%m/%d/%y %H:%M:%S,") 
          #  logstr += ("%4.1f,%5.1f,%4.1f" % (g_lv, g_hv ,g_temp) )
          #  logstr += (",%6.1f" % (tsession) )
          #  logstr += (",%6.1f,%6.1f,%7.1f" % ( g_stat_tracker.wh_drive.integrator/3600.0, g_stat_tracker.wh_regen.integrator/3600.0, g_stat_tracker.wh_net.integrator/3600.0 ) )
          #  logstr += (",%6.1f,%6.1f,%5.1f" % ( g_stat_tracker.wh_ac.integrator/3600.0, g_stat_tracker.wh_hv.integrator/3600.0, g_stat_tracker.wh_lv.integrator/3600.0 ) )

            l_wh_drive=g_stat_tracker.wh_drive.integrator/3600.0
            l_wh_regen=g_stat_tracker.wh_regen.integrator/3600.0
            l_wh_net=g_stat_tracker.wh_net.integrator/3600.0
            l_wh_ac=g_stat_tracker.wh_ac.integrator/3600.0
            l_wh_hv=g_stat_tracker.wh_hv.integrator/3600.0
            l_wh_lv=g_stat_tracker.wh_lv.integrator/3600.0
            l_wh_pv=g_stat_tracker.wh_pv.integrator/3600.0
            l_t_ci=g_stat_tracker.inlet_temperature.last_value
            l_t_co=g_stat_tracker.outlet_temperature.last_value
            l_v_twelve=g_stat_tracker.twelve_voltage.last_value
            l_v_stack=g_stat_tracker.pack_voltage.last_value
            l_v_sum=g_stat_tracker.sum_cell()
            l_v_min=g_stat_tracker.min_cell()
            l_v_max=g_stat_tracker.max_cell()
            l_t_avg=g_stat_tracker.avg_cell_temperature()
            l_t_min=g_stat_tracker.min_cell_temperature()
            l_t_max=g_stat_tracker.max_cell_temperature()
            l_i_pv=g_stat_tracker.a_pv.last_value
            l_t_fet=g_stat_tracker.fet_temperature.last_value
            l_t_trans=g_stat_tracker.trans_temperature.last_value
            l_w_pv=g_stat_tracker.wh_pv.last_value
            l_wh_pv_nv=g_stat_tracker.wh_pv_nv.last_value
            l_msfc=g_stat_tracker.msfc.last_value
            l_msfc_n=g_stat_tracker.msfc.count
            l_v_pv=g_stat_tracker.pv_voltage.last_value
            l_buck_current=g_stat_tracker.buck_current.last_value
            l_buck_voltage=g_stat_tracker.buck_voltage.last_value
            l_buck_temperature=g_stat_tracker.buck_temperature.last_value
            l_v_pcdb=g_stat_tracker.pcdb_vhv.last_value
            l_str =  "1"
            l_str+=  datetime.datetime.now().strftime(",%Y-%m-%dT%H:%M:%S") 
            l_str+= ",%6.1f" % l_wh_drive
            l_str+= ",%6.1f" % l_wh_regen
            l_str+= ",%7.1f" % l_wh_net
            l_str+= ",%6.1f" % l_wh_ac
            l_str+= ",%6.1f" % l_wh_hv
            l_str+= ",%6.1f" % l_wh_lv
            l_str+= ",%6.1f" % l_wh_pv
            l_str+= ",%1.0f" % 0 #l_s_drive
            l_str+= ",%1.0f" % 0 #l_s_ac
            l_str+= ",%1.0f" % 0 #l_s_pv
            l_str+= ",%5.0f" % tsession #l_s_wake
            l_str+= ",%5.2f" % l_v_twelve
            l_str+= ",%5.1f" % l_v_stack
            l_str+= ",%7.3f" % l_v_sum
            l_str+= ",%5.3f" % l_v_min
            l_str+= ",%5.3f" % l_v_max
            l_str+= ",%4.1f" % l_t_ci
            l_str+= ",%4.1f" % l_t_co
            l_str+= ",%4.1f" % l_t_avg
            l_str+= ",%4.1f" % l_t_min
            l_str+= ",%4.1f" % l_t_max
            l_str+= ",%1.0f" % 0 #l_m_dist
            l_str+=",%s,%s,%s,%s" % ( g_ghi, g_ghy, g_gti, g_gty )
            l_str+= ",%4.1f" % l_t_fet
            l_str+= ",%4.1f" % l_t_trans
            l_str+= ",%5.3f" % l_i_pv
            l_str+= ",%5.1f" % l_wh_pv_nv
            l_str+= ",%5.1f" % l_w_pv
            l_str+= ",%3.1f" % l_msfc
            l_str+= ",%3.0f" % l_msfc_n
            l_str+= ",%5.2f" % l_v_pv
            l_str+= ",%6.2f" % l_buck_current
            l_str+= ",%6.2f" % l_buck_voltage
            l_str+= ",%3.0f" % l_buck_temperature
            l_str+= ",%6.2f" % l_v_pcdb

            #print( l_str )
            with open("/home/pi/volt/energylog.csv","a") as f:
                f.write(l_str+"\n")
           # os.system("sync ")


        if string_daily:
            string_daily=0
            d_str =  "2"    # 0  string type
            d_str+=  datetime.datetime.now().strftime(",%Y-%m-%dT%H:%M:%S") # 1 timestamp
            d_str+= ",%6.1f" % d_wh_drive  #  2
            d_str+= ",%6.1f" % d_wh_regen  #  3
            d_str+= ",%7.1f" % d_wh_net    #  4
            d_str+= ",%6.1f" % d_wh_ac     #  5
            d_str+= ",%6.1f" % d_wh_hv     #  6
            d_str+= ",%6.1f" % d_wh_lv     #  7
            d_str+= ",%6.1f" % d_wh_pv     #  8
            d_str+= ",%5.0f" % d_s_drive   #  9
            d_str+= ",%5.0f" % d_s_ac      # 10
            d_str+= ",%5.0f" % d_s_pv      # 11 
            d_str+= ",%5.0f" % d_s_wake    # 12
            d_str+= ",%5.2f" % d_v_twelve  # 13
            d_str+= ",%5.1f" % d_v_stack   # 14
            d_str+= ",%7.3f" % d_v_sum     # 15
            d_str+= ",%5.3f" % d_v_min     # 16
            d_str+= ",%5.3f" % d_v_max     # 17
            d_str+= ",%4.1f" % d_t_ci      # 18
            d_str+= ",%4.1f" % d_t_co      # 19
            d_str+= ",%4.1f" % d_t_avg     # 20
            d_str+= ",%4.1f" % d_t_min     # 21
            d_str+= ",%4.1f" % d_t_max     # 22
            d_str+= ",%6.1f" % d_m_dist    # 23 

            d_str+=",%s,%s,%s,%s" % ( g_ghi, g_ghy, g_gti, g_gty )
            with open("/home/pi/volt/dailyforward.csv","a") as f:
                f.write(d_str+"\n")
            #print(d_str)


        if upload:
            upload=0
            # Uncomment this after adding your own website details 
            #os.system("tac /home/pi/volt/dailyforward.csv | curl --max-time 45 -T - -u username:password ftp://ftp.realsolarcars.com/v000001.csv")


        if reset_daily:
            d_wh_drive=0.0
            d_wh_regen=0.0
            d_wh_net=0.0
            d_wh_ac=0.0
            d_wh_hv=0.0
            d_wh_lv=0.0
            d_wh_pv=0.0
            d_s_pv=0.0
            d_s_drive=0.0
            d_s_ac=0.0
            d_s_wake=0.0
            d_m_dist=0.0
            d_v_twelve=0.0
            d_v_stack=0.0
            d_v_sum=0.0
            d_v_min=0.0
            d_v_max=0.0
            d_t_ci=88.8
            d_t_co=88.8
            d_t_avg=88.8
            d_t_min=88.8
            d_t_max=88.8
            reset_daily=0

        if reset_stats:
            reset_stats=0
            g_stat_tracker.reset_all()

        lastlooptimedate=thislooptimedate
        time.sleep(.1)

